//////////////////////////////////////////////////////////////////////////////////
// 
// SmileSoundDecoder Sketch Software
// Copyright(C)'2023- DesktopStation, Nagoden, Fujigaya2, Maison de DCC
// 
// 本ソフトウェアは、上記の著作者によって開発されたものであり、利用には
// ライセンス契約が必要となります。
// 
//////////////////////////////////////////////////////////////////////////////////



//#define DEBUG      //リリースのときはコメントアウトすること


#include <arduino.h>
#include <hardware/adc.h>
#include <hardware/clocks.h>
#include <hardware/pll.h>
#include <hardware/watchdog.h>
#include <hardware/sync.h>

#include "pico/multicore.h"
#include "hardware/irq.h"
#include "NmraDcc_pio.h"
#include "dcc_pulse_dec.h"


// Min and max valid packet lengths
#define kPACKET_LEN_MIN               3
#define kPACKET_LEN_MAX               6


//DCC PINS

#define PIN_DCC				2


#define DECODER_ADDRESS 3


typedef struct
{
    int count;
    byte validBytes;
    byte data[6];
} DCCPacket;

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// The dcc decoder object and global data
//
int gPacketCount = 0;
int gIdlePacketCount = 0;
int gLongestPreamble = 0;

DCCPacket gPackets[25];

//各種設定、宣言

//Programmable IO (PIO)
PIO gPio;
uint gPio_SM;
unsigned long gRecvPulseWidth;


//使用クラスの宣言
NmraDcc   Dcc;
DCC_MSG  Packet;


unsigned long gPreviousL1 = 0;
unsigned long gPreviousL5 = 0;


uint8_t gButtonStatus = 0;

uint16_t gCount_Err = 0;
uint16_t gCount_Idle = 0;
uint16_t gCount_Reset = 0;
uint16_t gCount_PreMin = 255;


void(* resetFunc) (void) = 0;  //declare reset function at address 0
uint16_t ReadEdcVoltage();

void decodeLocoPacket(uint8_t inPacket1, uint8_t inPacket2);


//------------------------------------------------------------------
// Arduino固有の関数 setup() :初期設定
//------------------------------------------------------------------
void setup()
{
	//PIO setup
	PIODCC_init();
	
	Serial.begin(115200);
	
	Serial.println("--------------------------------------");
	Serial.println("Desktop Station Monitor               ");
	Serial.println("--------------------------------------");
	Serial.println("100 Ready");  
	Serial.println("@DSG,999,");

	
	//ファンクションの割り当てピン初期化
	Dccinit();
	
	

	//Reset task
	gPreviousL5 = millis();	
	
}

//------------------------------------------------------------------
// DCC初期化処理）
//------------------------------------------------------------------
void Dccinit(void)
{
	
	// Call the main DCC Init function to enable the DCC Receiver
	Dcc.init( MAN_ID_DESKTOPSTATION, 1,   FLAGS_MY_ADDRESS_ONLY , 0 );
	
}

/**
* @brief PIOの初期設定
* @return なし
* @details PIOにDCC受信アセンブリコードをダウンロードして実行開始します。受信割り込みも設定します。
*/

void PIODCC_init()
{
	
	//DCC pulse input
    gpio_init(PIN_DCC);
	gpio_set_dir(PIN_DCC, false);
	gpio_disable_pulls(PIN_DCC);

	//PIOのch 0を確保
    gPio = pio0;

    //アセンブラのPIOプログラムへのアドレス
    uint offset = pio_add_program(gPio, &dcc_pulse_dec_program);

    // 空いているステートマシンの番号を取得
    gPio_SM = pio_claim_unused_sm(gPio, true);
	
	
	irq_set_exclusive_handler(PIO0_IRQ_0, dccrecvIrqHandler);
	irq_set_enabled(PIO0_IRQ_0, true);
	irq_set_priority(PIO0_IRQ_0, PICO_DEFAULT_IRQ_PRIORITY );
	irq_clear(PIO0_IRQ_0) ;	     				   // クリア
	irq_set_enabled(PIO0_IRQ_0, true) ;			   // 許可
	
	// PIOプログラムを初期化
	dcc_pulse_dec_init(gPio, gPio_SM, offset, PIN_DCC);
	
}

/**
* @brief PIOクリア
* @return なし
* @details PIOのFIFOをクリアします。
*/

void PIODCC_Flush()
{
	gRecvPulseWidth = pio_sm_get(gPio, gPio_SM);
}

/**
* @brief PIODCC受信割り込み
* @return なし
* @details DCCパルスのFIFOが溜まった時に発生する受信割り込み処理
*/

void dccrecvIrqHandler()
{
	int aTimeOutCheckCnt = 1;
	int aFIFOstackCnt = 0;

	// clear interrupt first
	//pio_interrupt_clear(gPio, 0);
	
	if( !pio_sm_is_rx_fifo_empty(gPio, gPio_SM))
	{
		
		//PIO RX FIFO check
		//aFIFOstackCnt = pio_sm_get_rx_fifo_level(gPio, gPio_SM);
		gRecvPulseWidth = pio_sm_get(gPio, gPio_SM);
		
		for( int j = 32 - ALIGNMENT_DCCBIT; j < 32; j++)
		{
			PushDCCbit((gRecvPulseWidth >> j) & 1);
		}
		
		//DCC receive process
		Dcc.process();
	}
}

//---------------------------------------------------------------------
// Arduino main loop (Core 0)
//---------------------------------------------------------------------

void loop()
{
	
	if( (millis() - gPreviousL5) >= 2000)
	{
		//Main Task
		DumpAndResetTable();
		
		//Reset task
		gPreviousL5 = millis();
		
	}
}

void DumpAndResetTable()
{
    char buffer60Bytes[60];
    
    Serial.print("Total Packet Count: ");
    Serial.println(gPacketCount, DEC);
    
    Serial.print("Idle Packet Count:  ");
    Serial.println(gIdlePacketCount, DEC);
        
    Serial.print("Longest Preamble:  ");
    Serial.println(gLongestPreamble, DEC);
    
    Serial.print("Error Packet Count:  ");
    Serial.println(gCount_Err, DEC);
	
    Serial.println("Count    Packet_Data");
    for( int i=0; i<(int)(sizeof(gPackets)/sizeof(gPackets[0])); ++i )
    {
        if( gPackets[i].validBytes > 0 )
        {
            Serial.print(gPackets[i].count, DEC);
            if( gPackets[i].count < 10 )
            {
                Serial.print("        ");
            }else{
                if( gPackets[i].count < 100 )
                {
                    Serial.print("       ");
                }else{
                    Serial.print("      ");
                }
            }
            Serial.println( MakePacketString(buffer60Bytes, gPackets[i].validBytes, &gPackets[i].data[0]) );
        }
        gPackets[i].validBytes = 0;
        gPackets[i].count = 0;
    }
    Serial.println("============================================");
    
    gPacketCount = 0;
    gIdlePacketCount = 0;
    gLongestPreamble = 0;
	gCount_Err = 0;
}

/* 全DCCメッセージを受信 */
void notifyDccMsg( DCC_MSG * Msg)
{
	uint8_t aErrorbyte = 0;
	byte packetBytes[6];
	byte byteCount;
	
	
	++gPacketCount;

	/* エラーチェック */
	for(uint8_t i = 0; i < Msg->Size; i++)
	{
		aErrorbyte = aErrorbyte ^ Msg->Data[i];
	}
	
	if( aErrorbyte != 0)
	{
		if( gCount_Err <= 65535)
		{
			gCount_Err++;
		}
		
		return;
	}
	else
	{
		//gWatchMode = 1;
	}
	
	/* プリアンブル数計算 */
	if( Msg->PreambleBits < gCount_PreMin)
	{
		gCount_PreMin = Msg->PreambleBits;
	}
	else if( Msg->PreambleBits > gLongestPreamble)
	{
		gLongestPreamble = Msg->PreambleBits;
	}
	
	
	for( int j = 0; j < Msg->Size; j++)
	{
		packetBytes[j] = Msg->Data[j];
	}
	
	byteCount = Msg->Size;
	
	// Walk table and look for a matching packet
	for( int i=0; i<(int)(sizeof(gPackets)/sizeof(gPackets[0])); ++i )
	{
		if( gPackets[i].validBytes )
		{
				// Not an empty slot. Does this slot match this packet? If so, bump count.
			if( gPackets[i].validBytes==byteCount )
			{
				char isPacket = true;
				for( int j=0; j<byteCount; j++)
				{
					if( gPackets[i].data[j] != packetBytes[j] )
					{
						isPacket = false;
						break;
					} 
				}
				if( isPacket )
				{
				   gPackets[i].count++;
				   return;
				}
			}
		}else{
				// Empty slot, just copy over data
			gPackets[i].count++;
			gPackets[i].validBytes = byteCount;
			for( int j=0; j<byteCount; j++)
			{
				gPackets[i].data[j] = packetBytes[j];
			}
			return;
		}
	}  
	
}

char* MakePacketString(char* buffer60Bytes, byte byteCount, byte* packet)
{
    buffer60Bytes[0] = 0;
    if( byteCount>=kPACKET_LEN_MIN && byteCount<=kPACKET_LEN_MAX )
    {
        int i = 0;
        for(byte byt=0; byt<byteCount; ++byt)
        {
            byte bit=0x80;
            while(bit)
            {
                buffer60Bytes[i++] = (packet[byt] & bit) ? '1' : '0';
                bit=bit>>1;
            }
            buffer60Bytes[i++] = ' ';
        }
        buffer60Bytes[--i] = 0;
    }
    return buffer60Bytes;
}
